-------------------------------------------------------------------------------
-- Blizzard Finnegan
-- 8-bit 2 function calc w/ double-dabble [top-level design]
-------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity top is
  port(
    clk             : in  std_logic;
    reset           : in  std_logic;
    input_switches  : in  std_logic_vector(7 downto 0);
    state_cycle     : in  std_logic;
    state_indicator : out std_logic_vector(3 downto 0);
    hundreds_disp   : out std_logic_vector(6 downto 0);
    tens_disp       : out std_logic_vector(6 downto 0);
    ones_disp       : out std_logic_vector(6 downto 0)
  );
end top;

architecture behaviour of top is

component double_dabble is
  port (
    result_padded           : in  std_logic_vector(11 downto 0); 
    ones                    : out std_logic_vector(3 downto 0);
    tens                    : out std_logic_vector(3 downto 0);
    hundreds                : out std_logic_vector(3 downto 0)
  );  
end component;

component generic_adder_beh is
  generic (
    bits       : integer := 4
  );
  port (
    a           : in  std_logic_vector(bits-1 downto 0); 
    b           : in  std_logic_vector(bits-1 downto 0);
    cin         : in std_logic;
    sum         : out std_logic_vector(bits-1 downto 0);
    cout        : out std_logic
  );  
end component;  

component double_buffer is
  generic (
    bits : integer := 3
  );
  port(
    d     : in  std_logic_vector(bits - 1 downto 0);
    clk   : in  std_logic;
    reset : in  std_logic;
    q     : out std_logic_vector(bits - 1 downto 0)
  );
end component;

component seven_seg is
  port (
    clk             : in  std_logic; 
    reset           : in  std_logic;
    bcd             : in std_logic_vector(3 downto 0);
    seven_seg_out   : out std_logic_vector(6 downto 0)
  );
end component;

signal a_store, b_store, show_add, show_sub, add_carry, sub_carry, state_cycle_z, state_cycle_zz, state_cycle_zzz, state_cycle_en : std_logic;
signal count_out: std_logic_vector(1 downto 0);
signal ones, tens, hundreds : std_logic_vector(3 downto 0);
signal input_zz : std_logic_vector(7 downto 0);
signal a, b, sub_b, add_out, sub_out: std_logic_vector(8 downto 0);
signal padded_add, padded_sub, padded_input, mux_disp: std_logic_vector(11 downto 0);

begin

padded_add <= "000" & add_out;
padded_sub <= "000" & sub_out;
padded_input <= "0000" & input_zz;
sub_b <= not b;
state_indicator <= ( 0 => a_store,
                     1 => b_store,
                     2 => show_add,
                     3 => show_sub);
state_cycle_en <= state_cycle_zzz and (not state_cycle_zz);

a_buffer: process(clk,a_store,reset)
begin
  if (reset = '1') then
    a <= ( others => '0' );
  elsif (rising_edge(clk)) then
    if ( a_store = '1' ) then
      a <= '0' & input_zz;
    end if;
  end if;
end process;

b_buffer: process(clk,b_store,reset)
begin
  if (reset = '1') then
    b <= ( others => '0' );
  elsif (rising_edge(clk)) then
    if ( b_store = '1' ) then
      b <= '0' & input_zz;
    end if;
  end if;
end process;

input_buffer: double_buffer
  generic map (
    bits => 8
  )
  port map(
    d => input_switches,
    q => input_zz,
    clk => clk,
    reset => reset
  );

adder: generic_adder_beh
  generic map(
    bits => 9
  )
  port map(
    a => a,
    b => b,
    cin => '0',
    sum => add_out,
    cout => add_carry
  );

subtractor: generic_adder_beh
  generic map(
    bits => 9
  )
  port map(
    a => a,
    b => sub_b,
    cin => '1',
    sum => sub_out,
    cout => sub_carry
  );

state_cycle_buffer: process(clk,state_cycle)
begin
  if (reset = '1') then
    state_cycle_z <= '0';
    state_cycle_zz <= '0';
    state_cycle_zzz <= '0';
  elsif rising_edge(clk) then
    state_cycle_z <= state_cycle;
    state_cycle_zz <= state_cycle_z;
    state_cycle_zzz <= state_cycle_zz;
  end if;
end process;

counter: process(state_cycle_en)
begin
  if(reset = '1') then
    count_out <= (others => '0');
  elsif (rising_edge(state_cycle_en)) then
    count_out <= std_logic_vector(unsigned(count_out) + 1);
  end if;
end process;

selector: process(count_out)
begin
  case count_out is
    when "00" => 
      a_store <= '1';
      b_store <= '0';
      show_add <= '0';
      show_sub <= '0';
    when "01" => 
      a_store <= '0';
      b_store <= '1';
      show_add <= '0';
      show_sub <= '0';
    when "10" => 
      a_store <= '0';
      b_store <= '0';
      show_add <= '1';
      show_sub <= '0';
    when "11" => 
      a_store <= '0';
      b_store <= '0';
      show_add <= '0';
      show_sub <= '1';
    when others =>
      a_store <= '1';
      b_store <= '0';
      show_add <= '0';
      show_sub <= '0';
  end case;
end process;

mux: process(a_store, b_store, show_add, show_sub)
begin
  if (a_store = '1' or b_store = '1') then
    mux_disp <= padded_input;
  elsif (show_add = '1') then
    mux_disp <= padded_add;
  elsif (show_sub = '1') then
    mux_disp <= padded_sub;
  end if;
end process;

bcd: double_dabble
  port map(
    result_padded => mux_disp,
    ones => ones,
    tens => tens,
    hundreds => hundreds
  );

show_hundreds: seven_seg
  port map (
    clk  => clk,
    reset => reset,
    bcd  => hundreds,
    seven_seg_out => hundreds_disp
  );

show_tens: seven_seg
  port map (
    clk  => clk,
    reset => reset,
    bcd  => tens,
    seven_seg_out => tens_disp
  );

show_ones: seven_seg
  port map (
    clk  => clk,
    reset => reset,
    bcd  => ones,
    seven_seg_out => ones_disp
  );
end behaviour;
